import unicodedata


def sanitize_char(char):
    """replace char with a dot, if it's a control or whitespace char

    Close to what hexdump does, but Unicode-aware. Characters that unicodedata
    is not aware of will also be replaced with a dot.

    >>> sanitize_char('A')
    'A'
    >>> sanitize_char('\U0001F678')
    '\U0001F678'
    >>> sanitize_char(' ')
    '.'
    >>> sanitize_char('\x01')
    '.'
    >>> # un-set Unicode character, should have category "Cn"
    >>> sanitize_char('\U000D0000')
    '.'
    >>> sanitize_char('a1')
    Traceback (most recent call last):
        ...
    TypeError: category() argument must be a unicode character, not str
    """
    category = unicodedata.category(char)
    if category[0] in ('C', 'Z'):
        return '.'
    return char


def print_line(line, env):
    """
    >>> from unidump.env import env
    >>> print_line([0, ['00A0', '00B0', '00C0'], 'ABC'], env(linelength=4))
          0    00A0 00B0 00C0         ABC\n
    >>> print_line([12, ['00A0', '1F678', '00C0'], 'A\U0001F678C'],
    ...            env(linelength=4))
         12    00A0 1F678 00C0        A\U0001F678C\n
    """
    env.output.write(env.lineformat.format(
        byte=line[0],
        repr=' '.join(line[1]).ljust(env.linelength*5-1),
        data=line[2]
    ))


def fill_and_print(current_line, byteoffset, representation, char, env):
    """
    >>> from unidump.env import env
    >>> _env = env(linelength=2)
    >>> current_line = [0, [], '']
    >>> current_line = fill_and_print(current_line, 0, '0076', 'v', _env)
    >>> current_line == [0, ['0076'], 'v']
    True
    >>> current_line = fill_and_print(current_line, 1, '0076', 'v', _env)
    >>> current_line = fill_and_print(current_line, 2, '0077', 'w', _env)
          0    0076 0076    vv\n
    >>> current_line == [1, ['0077'], 'w']
    True
    """
    if len(current_line[1]) >= env.linelength:
        print_line(current_line, env)
        current_line = [byteoffset, [representation], char]
    else:
        current_line[1].append(representation)
        current_line[2] += char

    return current_line



#!/usr/bin/env python3
"""
hexdump(1) for Unicode data
"""


from unidump.output import sanitize_char, print_line, fill_and_print


version = '1.1.1'


def unidump(inbytes, env):
    """take a list of bytes and print their Unicode codepoints

    >>> import io
    >>> from unidump.env import env
    >>> unidump(io.BytesIO(b'\\x01\\xF0\\x9F\\x99\\xB8ABC'), env(linelength=4))
          0    0001 1F678 0041 0042    .\U0001F678AB
          7    0043                   C
    >>> unidump(io.BytesIO(b'\\xD7'), env(linelength=4))
          0    ?D7?                   X
    >>> unidump(io.BytesIO(b'\\xD7'), env(linelength=4, encoding='latin1'))
          0    00D7                   \u00D7
    """

    byteoffset = 0
    bytebuffer = b''
    current_line = [0, [], '']

    byte = inbytes.read(1)
    while byte:
        byteoffset += 1
        bytebuffer += byte

        try:
            char = bytebuffer.decode(env.encoding)
        except UnicodeDecodeError:
            next_byte = inbytes.read(1)
            if not next_byte or len(bytebuffer) >= 4:
                for i, x in enumerate(bytebuffer):
                    current_line = (
                        fill_and_print(current_line, byteoffset - 4 + i,
                                       '?{:02X}?'.format(x), 'X', env)
                    )
                bytebuffer = b''
            byte = next_byte
            continue
        else:
            current_line = (
                fill_and_print(current_line, byteoffset - len(bytebuffer),
                               '{:04X}'.format(ord(char)), sanitize_char(char),
                               env)
            )

        bytebuffer = b''
        byte = inbytes.read(1)

    print_line(current_line, env)



import sys


class env(object):

    linelength = 16

    encoding = 'utf-8'

    lineformat = '{byte:>7}    {repr}    {data}\n'

    output = sys.stdout

    def __init__(self, linelength=None, encoding=None, lineformat=None):
        if linelength is not None:
            self.linelength = linelength
        if encoding is not None:
            self.encoding = encoding
        if lineformat is not None:
            self.lineformat = lineformat



import argparse
import sys
from unidump import version, unidump
from unidump.env import env


description = '''  A Unicode codepoint dump.

  Think of it as  hexdump(1)  for Unicode.  The command analyses  the input and
  prints then three columns:  the raw byte count of the first codepoint in this
  row,  codepoints in their hex notation,  and finally the raw input characters
  with control and whitespace replaced by a dot.

  Invalid byte sequences are represented with an “X” and with the hex value en-
  closed in question marks, e.g., “?F5?”.

  You can pipe in  data from stdin,  select several files at once,  or even mix
  all those input methods together.'''

epilog = '''Examples:

* Basic usage with stdin:

      echo -n 'ABCDEFGHIJKLMNOP' | unidump -n 4
            0    0041 0042 0043 0044    ABCD
            4    0045 0046 0047 0048    EFGH
            8    0049 004A 004B 004C    IJKL
           12    004D 004E 004F 0050    MNOP

* Dump the code points translated from another encoding:

      unidump -c latin-1 some-legacy-file

* Dump many files at the same time:

      unidump foo-*.txt

* Control characters and whitespace are safely rendered:

      echo -n -e '\\x01' | unidump -n 1
           0    0001    .

* Finally learn what your favorite Emoji is composed of:

      ( echo -n -e '\\xf0\\x9f\\xa7\\x9d\\xf0\\x9f\\x8f\\xbd\\xe2' ; \\
        echo -n -e '\\x80\\x8d\\xe2\\x99\\x82\\xef\\xb8\\x8f' ; ) | \\
      unidump -n 5
           0    1F9DD 1F3FD 200D 2642 FE0F    .🏽.♂️

  See  <http://emojipedia.org/man-elf-medium-skin-tone/> for images.  The “elf”
  emoji (the first character) is replaced with a dot here,  because the current
  version of Python’s unicodedata doesn’t know of this character yet.

* Use it like strings(1):

      unidump -e '{data}' some-file.bin

  This will replace  every unknown byte from the input file  with “X” and every
  control and whitespace character with “.”.

* Only print the code points of the input:

      unidump -e '{repr}'$'\\n' -n 1 some-file.txt

  This results in  a stream of codepoints in hex notation,  each on a new line,
  without byte counter  or rendering of actual data.  You can use this to count
  the total amount of characters  (as opposed to raw bytes)  in a file,  if you
  pipe it through `wc -l`.
'''


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog='unidump',
        description=description,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument('files', nargs='*', metavar='FILE', default=('-',),
                        help='input files. Use `-\' or keep empty for stdin.')
    parser.add_argument('-n', '--length', type=int, default=16,
                        dest='linelength', metavar='LENGTH',
                        help='format output using this much input characters. '
                        'Default is %(default)s characters.')
    parser.add_argument('-c', '--encoding', type=str, default='utf-8',
                        metavar='ENC',
                        help='interpret input in this encoding. Default is '
                        '%(default)s. You can choose any encoding that Python '
                        'supports, e.g. “latin-1”.')
    parser.add_argument('-e', '--format', type=str, default=None,
                        dest='lineformat', metavar='FORMAT',
                        help=(
                            'specify a custom format in Python’s {} notation. '
                            'Default is “%s”. '
                            'See examples below on how to use this option.'
                        ) % env.lineformat.replace('\n', '\\n'))
    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s {}'.format(version))

    a = parser.parse_args(args)

    try:
        for filename in a.files:
            if filename == '-':
                infile = sys.stdin.buffer
            else:
                try:
                    infile = open(filename, 'rb')
                except FileNotFoundError:
                    sys.stdout.flush()
                    sys.stderr.write('File {} not found.\n'.format(filename))
                    continue
                except IsADirectoryError:
                    sys.stdout.flush()
                    sys.stderr.write('{} is a directory.\n'.format(filename))
                    continue
            unidump(infile, env=env(linelength=a.linelength, encoding=a.encoding, lineformat=a.lineformat))
    except KeyboardInterrupt:
        sys.stdout.flush()
        return 1
    else:
        return 0



from setuptools import setup, find_packages
from codecs import open
from os import path

from unidump import version
from unidump.cli import description, epilog


long_description = description + '\n\n' + epilog

setup(
    name='unidump',
    version=version,

    description='hexdump for your unicode data',
    long_description=long_description,

    url='https://github.com/Codepoints/unidump',

    author='Manuel Strehl',
    author_email='boldewyn@gmail.com',

    license='MIT',

    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Debuggers',
        'Topic :: Software Development :: Internationalization',
        'Topic :: Text Processing :: General',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    keywords='unicode hexdump codepoint utility',

    packages=['unidump'],

    entry_points={
        'console_scripts': [
            'unidump=unidump.cli:main',
        ],
    },
)
